/*
 *  knockd.c
 * 
 *  Copyright (c) 2004 by Judd Vinet <jvinet@zeroflux.org>
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <net/if.h>
#include <bits/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <getopt.h>
#include <syslog.h>
#include <pcap.h>
#include "list.h"

static char version[] = "0.2.1";

#define PROTO_ICMP  1
#define PROTO_TCP   6
#define PROTO_UDP  17

#define SEQ_TIMEOUT 25 /* default knock timeout in seconds */
#define SEQ_MAX     32 /* maximum number of ports in a knock sequence */

/* knock/event tuples */
typedef struct opendoor {
	char name[128];
	unsigned short seqcount;
	unsigned short sequence[SEQ_MAX];
	unsigned short protocol;
	time_t timeout;
	char *cmd;
	unsigned short flag_fin;
	unsigned short flag_syn;
	unsigned short flag_rst;
	unsigned short flag_psh;
	unsigned short flag_ack;
	unsigned short flag_urg;
} opendoor_t;
PMList *doors = NULL;

/* we keep one list of knock attempts, one per IP address,
 * and increment the stage as they progress through the sequence.
 */
typedef struct knocker {
	opendoor_t *door;
	short stage;
	char src[16];
	time_t seq_start;
} knocker_t;
PMList *attempts = NULL;

/* function prototypes */
void dprint(char *fmt, ...);
void vprint(char *fmt, ...);
void logprint(char *fmt, ...);
void cleanup(int signum);
void child_exit(int signum);
void ver();
void usage();
char* strtoupper(char *str);
char* trim(char *str);
void runCommand(char *cmd);
int parseconfig(char *configfile);
int get_mac(const char* iface, u_char *addr);
void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet);

pcap_t *cap = NULL;
FILE *logfd = NULL;
u_char if_mac[6];

int  o_usesyslog = 0;
int  o_verbose   = 0;
int  o_debug     = 0;
int  o_daemon    = 0;
char o_int[32]           = "eth0";
char o_cfg[PATH_MAX]     = "/etc/knockd.conf";
char o_pidfile[PATH_MAX] = "/var/run/knockd.pid";
char o_logfile[PATH_MAX] = "";

int main(int argc, char **argv)
{
	char pcapErr[PCAP_ERRBUF_SIZE] = "";
	int opt, ret, optidx = 1;
	static struct option opts[] =
	{
		{"verbose",   no_argument,       0, 'v'},
		{"debug",     no_argument,       0, 'D'},
		{"daemon",    no_argument,       0, 'd'},
		{"interface", required_argument, 0, 'i'},
		{"config",    required_argument, 0, 'c'},
		{"help",      no_argument,       0, 'h'},
		{"version",   no_argument,       0, 'V'},
		{0, 0, 0, 0}
	};

	while((opt = getopt_long(argc, argv, "vDdi:c:hV", opts, &optidx))) {
		if(opt < 0) {
			break;
		}
		switch(opt) {
			case 0:   break;
			case 'v': o_verbose = 1; break;
			case 'D': o_debug = 1; break;
			case 'd': o_daemon = 1; break;
			case 'i': strncpy(o_int, optarg, sizeof(o_int)); break;
			case 'c': strncpy(o_cfg, optarg, sizeof(o_cfg)); break;
			case 'V': ver();
			case 'h': /* fallthrough */
			default: usage();
		}
	}

	if(parseconfig(o_cfg)) {
		exit(1);
	}
	if(o_usesyslog) {
		openlog("knockd", 0, LOG_USER);
	}
	if(strlen(o_logfile)) {
		/* open the log file */
		logfd = fopen(o_logfile, "a");
		if(logfd == NULL) {
			perror("warning: cannot open logfile");
		}
	}

	if(get_mac(o_int, if_mac)) {
		fprintf(stderr, "could not get MAC address for %s\n", o_int);
		exit(1);
	}

	cap = pcap_open_live(o_int, 65535, 0, 0, pcapErr);
	if(strlen(pcapErr)) {
		fprintf(stderr, "could not open %s: %s\n", o_int, pcapErr);
	}
	if(cap == NULL) {
		exit(1);
	}

	if(o_daemon) {
		if(fork() > 0) {
			/* parent */
			exit(0);
		} else {
			/* child */
			FILE *pidfp;
			setsid();
			close(0);			
			close(1);
			close(2);
			/* write our PID to the pidfile*/
			if((pidfp = fopen(o_pidfile, "w"))) {
				fprintf(pidfp, "%d\n", getpid());
				fclose(pidfp);
			}
		}
	}

	signal(SIGINT, cleanup);
	signal(SIGTERM, cleanup);
	signal(SIGCHLD, child_exit);

	vprint("listening on %s...\n", o_int);
	logprint("starting up, listening on %s", o_int);
	ret = 1;
	while(ret >= 0) {
		ret = pcap_dispatch(cap, -1, sniff, NULL);
	}
	dprint("bailed out of main loop! (ret=%d)\n", ret);
	pcap_perror(cap, "pcap");

	cleanup(0);
	/* notreached */
	exit(0);
}

void dprint(char *fmt, ...)
{
	va_list args;
	if(o_debug) {
		va_start(args, fmt);
		vprintf(fmt, args);
		va_end(args);
		fflush(stdout);
	}
}

void vprint(char *fmt, ...)
{
	va_list args;
	if(o_verbose) {
		va_start(args, fmt);
		vprintf(fmt, args);
		va_end(args);
		fflush(stdout);
	}
}

/* Output a message to syslog and/or a logfile */
void logprint(char *fmt, ...)
{
	char msg[1024];
	va_list args;
	va_start(args, fmt);
	vsnprintf(msg, 1024, fmt, args);
	va_end(args);
	if(o_usesyslog) {
		syslog(LOG_NOTICE, "%s", msg);
	}
	if(logfd) {
		time_t t;
		struct tm *tm;
		t = time(NULL);
		tm = localtime(&t);
		
		fprintf(logfd, "[%02d/%02d/%02d %02d:%02d] %s\n", tm->tm_mon+1, tm->tm_mday,
			tm->tm_year-100, tm->tm_hour, tm->tm_min, msg);
		fflush(logfd);
	}
}

/* Signal handlers */
void cleanup(int signum)
{
	vprint("closing...\n");
	logprint("shutting down");
	pcap_close(cap);
	if(o_daemon) {
		unlink(o_pidfile);
	}
	exit(signum);
}

void child_exit(int signum)
{
	/* child wants to exit, let em die */
	wait(NULL);
	return;
}

void usage() {
	printf("usage: knockd [options]\n");
	printf("options:\n");
	printf("  -i, --interface <int>  network interface to listen on (default \"eth0\")\n");
	printf("  -d, --daemon           run as a daemon\n");
	printf("  -c, --config <file>    use an alternate config file\n");
	printf("  -D, --debug            output debug messages\n");
	printf("  -v, --verbose          be verbose\n");
	printf("  -V, --version          display version\n");
	printf("  -h, --help             this help\n");
	printf("\n");
	exit(1);
}

void ver() {
	printf("knockd %s\n", version);
	printf("Copyright (C) 2004 Judd Vinet <jvinet@zeroflux.org>\n");
	exit(0);
}

/* Convert a string to uppercase
 */
char* strtoupper(char *str)
{
	char *ptr = str;

	while(*ptr) {
		(*ptr) = toupper(*ptr);
		ptr++;
	}
	return str;
}

/* Trim whitespace and newlines from a string
 */
char* trim(char *str)
{
	char *pch = str;
	while(isspace(*pch)) {
		pch++;
	}
	if(pch != str) {
		memmove(str, pch, (strlen(pch) + 1));
	}
	
	pch = (char*)(str + (strlen(str) - 1));
	while(isspace(*pch)) {
		pch--;
	}
	*++pch = '\0';

	return str;
}

/* Parse a config file
 */
int parseconfig(char *configfile)
{
	FILE *fp = NULL;
	char line[PATH_MAX+1];
	char *ptr = NULL;
	char *key = NULL;
	int linenum = 0;
	char section[256] = "";
	opendoor_t *door = NULL;
	PMList *lp;

	if((fp = fopen(configfile, "r")) == NULL) {
		perror(configfile);
		return(1);
	}

	while(fgets(line, PATH_MAX, fp)) {
		linenum++;
		trim(line);
		if(strlen(line) == 0 || line[0] == '#') {
			continue;
		}
		if(line[0] == '[' && line[strlen(line)-1] == ']') {
			/* new config section */
			ptr = line;
			ptr++;
			strncpy(section, ptr, sizeof(section));
			section[strlen(section)-1] = '\0';
			dprint("config: new section: '%s'\n", section);
			if(!strlen(section)) {
				fprintf(stderr, "config: line %d: bad section name\n", linenum);
				return(1);
			}
			if(strcmp(section, "options")) {
				/* start a new knock/event record */
				door = malloc(sizeof(opendoor_t));
				if(door == NULL) {
					perror("malloc");
					exit(1);
				}
				strncpy(door->name, section, sizeof(door->name));
				door->seqcount = 0;
				door->timeout  = SEQ_TIMEOUT; /* default timeout (seconds)  */
				door->protocol = PROTO_TCP;   /* default protocol */
				door->cmd = NULL;
				doors = list_add(doors, door);
			}
		} else {
			/* directive */
			if(!strlen(section)) {
				fprintf(stderr, "config: line %d: all directives must belong to a section\n", linenum);
				return(1);
			}
			ptr = line;
			key = strsep(&ptr, "=");
			if(key == NULL) {
				fprintf(stderr, "config: line %d: syntax error\n", linenum);
				return(1);
			}
			trim(key);
			key = strtoupper(key);
			if(ptr == NULL) {
				if(!strcmp(key, "USESYSLOG")) {
					o_usesyslog = 1;
					dprint("config: usesyslog\n");
				} else {
					fprintf(stderr, "config: line %d: syntax error\n", linenum);
					return(1);			
				}
			} else {
				trim(ptr);
				if(!strcmp(section, "options")) {
					if(!strcmp(key, "LOGFILE")) {
						strncpy(o_logfile, ptr, PATH_MAX);
						dprint("config: log file: %s\n", o_logfile);
					} else if(!strcmp(key, "PIDFILE")) {
						strncpy(o_pidfile, ptr, PATH_MAX);
						dprint("config: pid file: %s\n", o_pidfile);

					} else {
						fprintf(stderr, "config: line %d: syntax error\n", linenum);
						return(1);
					}
				} else {
					if(door == NULL) {
						fprintf(stderr, "config: line %d: \"%s\" can only be used within a Door section\n",
								linenum, key);
						return(1);
					}
					if(!strcmp(key, "SEQUENCE")) {
						char *num;
						int i;
						while((num = strsep(&ptr, ","))) {
							if(door->seqcount >= SEQ_MAX) {
								fprintf(stderr, "config: section %s: too many ports in knock sequence\n",
										door->name);
								return(1);
							}
							door->sequence[door->seqcount++] = (unsigned short)atoi(num);
						}
						dprint("config: %s: sequence: ", door->name);
						for(i = 0; i < door->seqcount; i++) {
							dprint((i == door->seqcount-1 ? "%u\n" : "%u,"), door->sequence[i]);
						}
					} else if(!strcmp(key, "PROTOCOL")) {
						if(!strcasecmp(ptr, "TCP")) {
							door->protocol = PROTO_TCP;
						} else if(!strcasecmp(ptr, "UDP")) {
							door->protocol = PROTO_UDP;
						} else {
							fprintf(stderr, "config: line %d: invalid protocol\n", linenum);
							return(1);
						}
						dprint("config: %s: protocol: %s\n", door->name, ptr);
					} else if(!strcmp(key, "TIMEOUT")) {
						door->timeout = (time_t)atoi(ptr);
						dprint("config: %s: timeout: %d\n", door->name, door->timeout);
					} else if(!strcmp(key, "COMMAND")) {
						door->cmd = malloc(sizeof(char) * (strlen(ptr)+1));
						if(door->cmd == NULL) {
							perror("malloc");
							exit(1);
						}
						strcpy(door->cmd, ptr);
						dprint("config: %s: cmd: %s\n", door->name, door->cmd);
					} else if(!strcmp(key, "TCPFLAGS")) {
						char *flag;
						strtoupper(ptr);
						while((flag = strsep(&ptr, ","))) {
							/* allow just some flags to be specified */
							if(!strcmp(flag,"FIN")) {
								door->flag_fin = 1;
							} else if(!strcmp(flag, "SYN")) {
								door->flag_syn = 1;
							} else if(!strcmp(flag, "RST")) {
								door->flag_rst = 1;
							} else if(!strcmp(flag, "PSH")) {
								door->flag_psh = 1;
							} else if(!strcmp(flag, "ACK")) {
								door->flag_ack = 1;
							} else if(!strcmp(flag, "URG")) {
								door->flag_urg = 1;
							} else {
								fprintf(stderr, "config: line %d: unrecognized flag \"%s\"\n",
										linenum, flag);
								return(1);
							}
							dprint("config: tcp flag: %s\n", flag);
						}
					} else {
						fprintf(stderr, "config: line %d: syntax error\n", linenum);
						return(1);
					}
				}
				line[0] = '\0';
			}
		}
	}
	fclose(fp);

	/* sanity checks */
	for(lp = doors; lp; lp = lp->next) {
		door = (opendoor_t*)lp->data;
		if(door->seqcount == 0) {
			fprintf(stderr, "error: section '%s' has an empty knock sequence\n", door->name);
			return(1);
		}
	}

	return(0);
}

/* Get the MAC address of an ethernet interface
 */
int get_mac(const char* iface, u_char *addr)
{
	struct ifreq ifr;
	struct ifreq *ifrptr;
	struct ifconf ifc;
	char buf[1024];
	int s, i;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if(s == -1) {
		return(-1);
	}

	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	ioctl(s, SIOCGIFCONF, &ifc);

	ifrptr = ifc.ifc_req;
	for(i = ifc.ifc_len / sizeof(struct ifreq); --i >= 0; ifrptr++) {
		if(!strcmp(ifrptr->ifr_name, iface)) {
			strcpy(ifr.ifr_name, ifrptr->ifr_name);
			if(ioctl(s, SIOCGIFFLAGS, &ifr) == 0) {
				if(!(ifr.ifr_flags & IFF_LOOPBACK)) {
					if(ioctl(s, SIOCGIFHWADDR, &ifr) == 0) {
						close(s);
						bcopy(ifr.ifr_hwaddr.sa_data, addr, 6);
						return(0);
					}
				}
			}
		}
	}
	return(1);
}

/* Sniff an interface, looking for port-knock sequences
 */
void sniff(u_char* arg, const struct pcap_pkthdr* hdr, const u_char* packet)
{
	int i;
	/* packet structs */
	struct ethhdr* eth = NULL;
	struct iphdr* ip   = NULL;
	struct tcphdr* tcp = NULL;
	struct udphdr* udp = NULL;
	char proto[8];
	/* TCP/IP data */	
	struct in_addr inaddr;
	unsigned short sport, dport;
	char srcIP[16], dstIP[16];
	/* timestamp */
	time_t pkt_secs = hdr->ts.tv_sec;
	struct tm* pkt_tm;
	char pkt_date[11];
	char pkt_time[9];
	PMList *lp;
	knocker_t *attempt = NULL;

	eth = (struct ethhdr*)packet;
	if(ntohs(eth->h_proto) != ETH_P_IP) {
		return;
	}
	/* make sure this packet was sent TO us, not sent BY us */
	for(i = 0; i < 6; i++) {
		if(eth->h_dest[i] != if_mac[i]) {
			return;
		}
	}
	
	ip = (struct iphdr*)(packet + sizeof(struct ethhdr));
	if(ip->version != 4) {
		/* no IPv6 yet */
		dprint("packet is not IPv4, ignoring...\n");
		return;
	}
	if(ip->protocol == PROTO_ICMP) {
		return;
	}
	sport = dport = 0;
	if(ip->protocol == PROTO_TCP) {
		strncpy(proto, "tcp", sizeof(proto));
		tcp = (struct tcphdr*)((u_char*)ip + (ip->ihl * 4));
		sport = ntohs(tcp->source);
		dport = ntohs(tcp->dest);
	}
	if(ip->protocol == PROTO_UDP) {
		strncpy(proto, "udp", sizeof(proto));
		udp = (struct udphdr*)((u_char*)ip + (ip->ihl * 4));
		sport = ntohs(udp->source);
		dport = ntohs(udp->dest);
	}

	/* if the incoming port is 22 (SSH) we ignore this packet.
	 * this makes it much easier to test over an SSH connection, otherwise
	 * the ssh traffic interferes with the knock
	 */
	/* commented out to test the FLAG detection
	if(dport == 22) return;*/

	/* get the date/time */
	pkt_tm = localtime(&pkt_secs);
	snprintf(pkt_date, 11, "%04d-%02d-%02d", pkt_tm->tm_year+1900, pkt_tm->tm_mon,
			pkt_tm->tm_mday);
	snprintf(pkt_time, 9, "%02d:%02d:%02d", pkt_tm->tm_hour, pkt_tm->tm_min,
			pkt_tm->tm_sec);

	/* convert IPs from binary to string */
	inaddr.s_addr = ip->saddr;
	strncpy(srcIP, inet_ntoa(inaddr), 16);
	inaddr.s_addr = ip->daddr;
	strncpy(dstIP, inet_ntoa(inaddr), 16);

	dprint("%s %s: %s: %s:%d -> %s:%d %d bytes\n", pkt_date, pkt_time,
			proto, srcIP, sport, dstIP, dport, hdr->len);

	/* clean up expired/completed/failed attempts */
	for(lp = attempts; lp; lp = lp->next) {
		int nix = 0;
		attempt = (knocker_t*)lp->data;
		if(attempt->stage >= attempt->door->seqcount) {
			dprint("removing successful knock attempt (%s)\n", attempt->src);
			nix = 1;
		}
		if(attempt->stage < 0) {
			dprint("removing failed knock attempt (%s)\n", attempt->src);
			nix = 1;
		}
		if(!nix && (pkt_secs - attempt->seq_start) >= attempt->door->timeout) {
			vprint("%s: %s: Timed Out (stage %d)\n", attempt->src, attempt->door->name, attempt->stage);
			logprint("%s: %s: Timed Out (stage %d)", attempt->src, attempt->door->name, attempt->stage);
			nix = 1;
		}
		if(nix) {
			/* splice this entry out of the list */
			if(lp->prev) lp->prev->next = lp->next;
			if(lp->next) lp->next->prev = lp->prev;
			if(lp == attempts) attempts = NULL;
			lp->prev = lp->next = NULL;
			list_free(lp);
			continue;
		}
	}

	attempt = NULL;
	/* look for this guy in our attempts list */
	for(lp = attempts; lp; lp = lp->next) {
		if(!strncmp(((knocker_t*)lp->data)->src, srcIP, sizeof(srcIP))) {
			attempt = (knocker_t*)lp->data;
			break;
		}
	}

	if(attempt) {
		int flagsmatch = 1;
		/* if tcp, check the flags to ignore the packets we don't want
		 * (don't even use it to cancel sequences)
		 */
		if(ip->protocol == PROTO_TCP) {
			if(attempt->door->flag_fin == 1 && tcp->fin != 1) {
				dprint("packet is not FIN, ignoring...\n");
				flagsmatch = 0;
			}
			if(attempt->door->flag_syn == 1 && tcp->syn != 1) {
				dprint("packet is not SYN, ignoring...\n");
				flagsmatch = 0;
			}
			if(attempt->door->flag_rst == 1 && tcp->rst != 1) {
				dprint("packet is not RST, ignoring...\n");
				flagsmatch = 0;
			}
			if(attempt->door->flag_psh == 1 && tcp->psh != 1) {
				dprint("packet is not PSH, ignoring...\n");
				flagsmatch = 0;
			}
			if(attempt->door->flag_ack == 1 && tcp->ack != 1) {
				dprint("packet is not ACK, ignoring...\n");
				flagsmatch = 0;
			}
			if(attempt->door->flag_urg == 1 && tcp->urg != 1) {
				dprint("packet is not URG, ignoring...\n");
				flagsmatch = 0;
			}
		}
		if(flagsmatch && ip->protocol == attempt->door->protocol &&
				dport == attempt->door->sequence[attempt->stage]) {
			/* level up! */
			attempt->stage++;
			vprint("%s: %s: Stage %d\n", attempt->src, attempt->door->name, attempt->stage);
			logprint("%s: %s: Stage %d", attempt->src, attempt->door->name, attempt->stage);
			if(attempt->stage >= attempt->door->seqcount) {
				vprint("%s: %s: OPEN SESAME\n", attempt->src, attempt->door->name);
				logprint("%s: %s: OPEN SESAME", attempt->src, attempt->door->name);
				if(attempt->door->cmd && strlen(attempt->door->cmd)) {
					/* run the associated command */
					if(fork() == 0) {
						/* child */
						int ret;
						char *ptr1, *ptr2;
						char origCmd[PATH_MAX];
						char parsedCmd[PATH_MAX] = "";
						setsid();
						/* replace all occurrences of %IP% with the source IP address */
						strncpy(origCmd, attempt->door->cmd, sizeof(origCmd));
						ptr1 = origCmd;
						while((ptr2 = strstr(ptr1, "%IP%"))) {
							ptr2[0] = '\0';
							strcat(parsedCmd, ptr1);
							strcat(parsedCmd, attempt->src);
							ptr1 = ptr2 + 4;
						}
						strcat(parsedCmd, ptr1);
						/* execute the parsed command via /bin/sh -c */
						logprint("running command: %s\n", parsedCmd);
						vprint("running command: %s\n", parsedCmd);
						ret = system(parsedCmd);
						if(ret == -1) {
							fprintf(stderr, "error: fork failed!\n");
							logprint("error: fork failed!");
						} else if(ret != 0) {
							fprintf(stderr, "%s: command returned non-zero status code (%d)\n",
									attempt->door->name, WEXITSTATUS(ret));
							logprint("%s: command returned non-zero status code (%d)",
									attempt->door->name, WEXITSTATUS(ret));
						}
						exit(0);
					}
				}
			}
		} else if(flagsmatch == 0) {
			/* TCP flags didn't match -- just ignore this packet, don't
			 * invalidate the knock.
			 */
		} else {
			/* invalidate the knock sequence, it will be removed in the
			 * next sniff() call.
			 */
			attempt->stage = -1;
		}
	} else {
		/* did they hit the first port correctly? */
		for(lp = doors; lp; lp = lp->next) {
			opendoor_t *door = (opendoor_t*)lp->data;
			/* if we're working with TCP, try to match the flags */
			if(ip->protocol == PROTO_TCP){
				if(door->flag_fin == 1 && tcp->fin != 1){dprint("packet is not FIN, ignoring...\n");continue;}
				if(door->flag_syn == 1 && tcp->syn != 1){dprint("packet is not SYN, ignoring...\n");continue;}
				if(door->flag_rst == 1 && tcp->rst != 1){dprint("packet is not RST, ignoring...\n");continue;}
				if(door->flag_psh == 1 && tcp->psh != 1){dprint("packet is not PSH, ignoring...\n");continue;}
				if(door->flag_ack == 1 && tcp->ack != 1){dprint("packet is not ACK, ignoring...\n");continue;}
				if(door->flag_urg == 1 && tcp->urg != 1){dprint("packet is not URG, ignoring...\n");continue;}
			}

			if(ip->protocol == door->protocol && dport == door->sequence[0]) {
				/* create a new entry */
				attempt = (knocker_t*)malloc(sizeof(knocker_t));
				if(attempt == NULL) {
					perror("malloc");
					exit(1);
				}
				strcpy(attempt->src, srcIP);
				attempt->stage = 1;
				attempt->seq_start = pkt_secs;
				attempt->door = door;
				vprint("%s: %s: Stage 1\n", attempt->src, door->name);
				logprint("%s: %s: Stage 1", attempt->src, door->name);
				attempts = list_add(attempts, attempt);
			}
		}
	}
}

/* vim: set ts=2 sw=2 noet: */
